#creates corpus of documents which have attached citations 
#!NOTE that some articles could be cited but they have no citations! 

import logging 
import leveldb
import csv

from nltk.tokenize import wordpunct_tokenize
from nltk.corpus import stopwords

from pubmed_parser import parse_article

logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.INFO)
logger = logging.getLogger('citations')

db = leveldb.LevelDB("/datastore/pubmed/db-cit") 
it = db.RangeIter(fill_cache=True)

corpus_csv = csv.writer(open("corpus/corpus_cit.csv", "w"))
author_csv = csv.writer(open("author_cit.csv", "w"))
citations_csv = csv.writer(open("citations.csv", "w"))

#=================== Gather keys of all articles with citations ====================
logger.info("==========Working on citations=============") 
cit = {} 
cc = 0 
for k, v in it:
  cit[k] = v 
  row = [k]
  row.extend(v.split(",")) 
  citations_csv.writerow(row) 
  
  cc += 1
  if cc % 100000 == 0: 
    logger.info(str(cc))  
    
print "len(cit)=", len(cit) 

#================== helper function for getting article abstract with citations =============
def get_with_cit(db_name): 
  logger.info("==========Working on " + db_name + " ================") 
  
  db_it = leveldb.LevelDB("/datastore/pubmed/" + db_name).RangeIter(fill_cache=True)
  
  count_k = 0 
  count_add = 0
  for k, v in db_it:
    count_k += 1
    
    if count_k % 10000 == 0: 
      logger.info("added=" + str(count_add) + " considered=" + str(count_k))
    
    if k not in cit:
      continue
    
    data = parse_article(v)
    
    abstract = data[1]
    authors = data[3] 
    
    if len(abstract) < 10:
      continue
    if len(authors) < 1:
      continue 
    
    tokens = [x.lower() for x in wordpunct_tokenize(abstract)]
    if len(tokens) < 10: 
      continue 
    
    row = [k] 
    row.extend([x.encode("utf8") for x in authors]) 
    author_csv.writerow(row) 
    
    row = [k]
    row.extend([x.encode("utf8") for x in tokens])
    corpus_csv.writerow(row) 
  
    count_add += 1
      
get_with_cit("db")
get_with_cit("db2")
get_with_cit("db3")
get_with_cit("db4")
get_with_cit("db5")
  