import logging
import cPickle

import csv
import re 
import sys 

import hashlib 

from gensim import corpora

#from nltk.stem.lancaster import LancasterStemmer

#======== FILTERING SETTINGS ============
NUM_DOCS = 400000000
MIN_WORD_LENGTH = 3 
MIN_DOC_LENGTH = 10 # safes some unexpected empty line errors 

# TODO implement
IS_STEM = False  

WORD_REGEXP = r'^[a-zA-Z]+[a-zA-Z0-9]*$'; # word must match WORD_REGEXP to be included

MAX_DOC_FREQ = 0.20; # if the word is in more than MAX_DOC_FREQ then it will be skipped 

# TODO implement 
POS_TAGS = ["C", "CD", "DT", "EX", "FW", "IN", "JJ", "JJR", "JJS", "LS", "MD", "NN", "NNS", "NNP", "NNPS", "PDT", "POS", "PRP", "PRP$", "RB", "RBR", "RBS", "RP", "SYM", "TO", "UH", "VB", "VBD", "VBG", "VBN", "VBP", "VBZ", "WDT", "WP", "WP$", "WRB"]

#=============== LOGGING ==================
db_name = sys.argv[1] 

logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.INFO)
logger = logging.getLogger('doc_used')

#===========================================================
#===========================================================
#=========================== SKIP_WORDS ====================
#===========================================================
#===========================================================

#=================== DOC_FREQUENCIES ==============
logger.info("ME: Loading doc frequencies") 
fdoc_freq = csv.reader(open("frequency/doc_"+db_name+".csv", "r"))
frequentwords = []
for wf in fdoc_freq: 
  if float(wf[0]) < MAX_DOC_FREQ: 
    continue 
  frequentwords.append(wf[1]) 

#=================== SKIP WORDS ============================
stopwords = set(["a", "about", "above", "after", "again", "against", "all", "am", "an", "and", "any", "are", "aren't", "as", "at", "be", "because", "been", "before", "being", "below", "between", "both", "but", "by", "can't", "cannot", "could", "couldn't", "did", "didn't", "do", "does", "doesn't", "doing", "don't", "down", "during", "each", "few", "for", "from", "further", "had", "hadn't", "has", "hasn't", "have", "haven't", "having", "he", "he'd", "he'll", "he's", "her", "here", "here's", "hers", "herself", "him", "himself", "his", "how", "how's", "i", "i'd", "i'll", "i'm", "i've", "if", "in", "into", "is", "isn't", "it", "it's", "its", "itself", "let's", "me", "more", "most", "mustn't", "my", "myself", "no", "nor", "not", "of", "off", "on", "once", "only", "or", "other", "ought", "our", "ours ", "ourselves", "out", "over", "own", "same", "shan't", "she", "she'd", "she'll", "she's", "should", "shouldn't", "so", "some", "such", "than", "that", "that's", "the", "their", "theirs", "them", "themselves", "then", "there", "there's", "these", "they", "they'd", "they'll", "they're", "they've", "this", "those", "through", "to", "too", "under", "until", "up", "very", "was", "wasn't", "we", "we'd", "we'll", "we're", "we've", "were", "weren't", "what", "what's", "when", "when's", "where", "where's", "which", "while", "who", "who's", "whom", "why", "why's", "with", "won't", "would", "wouldn't", "you", "you'd", "you'll", "you're", "you've", "your", "yours", "yourself", "yourselves"]) 

#TODO 
bullshitwords = set(["abstract", "publisher", "welcome", "letters", "article", "editor", "articles", "published", "publish", "editing", "author", "authors", "original", "reply", "previous", "publication", "publications", "appl", "opt", "appendix", "errors", "corrected"]) 

skipwords = set(list(stopwords) + list(bullshitwords) + list(frequentwords))

doc_hashes = set([]) 

logger.info("ME: StopWords("+str(len(stopwords))+")")
logger.info("ME: FrequentWords("+str(len(frequentwords))+")")
logger.info("ME: BullshitWords("+str(len(bullshitwords))+")")
logger.info("ME: =SkipWords("+str(len(skipwords))+")")

#=================== GENERATOR of the CORPUS =============
#  it also saves the transforms 
    
logger.info("ME: Loading raw corpus") 

fh_corpus_csv = open("corpus/" + db_name + ".csv", "r")
corpus_csv = csv.reader(fh_corpus_csv) 
docs_used_csv = csv.writer(open("model/docs_used_" + db_name + ".csv", "w"))

#st = LancasterStemmer()

word2id = {} # dictionary (reversed) 
b = 0
c = 0
for line in corpus_csv:
  #pos_tags = nltk.pos_tag(line)
  #print pos_tags
  
  b += 1 
  
  doc_to_use = [w for w in line if w not in skipwords and len(w) >= MIN_WORD_LENGTH and re.match(WORD_REGEXP, w)]
  
  if len(doc_to_use) < MIN_DOC_LENGTH:
    continue
  
  # checking uniqueness of the abstract 
  h = hashlib.sha224(" ".join(doc_to_use)).hexdigest()
  if h in doc_hashes:
    continue 
  doc_hashes.add(h) 
  
  #append id 
  row = [line[0]]
  row.extend(doc_to_use)
  docs_used_csv.writerow(row)
  
  #add to dictionary 
  for x in doc_to_use:
    if x not in word2id:
      word2id[x] = len(word2id) 
  
  c += 1
  if c % 10000 == 0: 
    logger.info("ME:   " + str(c)) 
  if c >= NUM_DOCS:
    break
    
logger.info("ME: BeforeCount="+str(b)+" AfterCount="+str(c))

#=================== SAVE DICTIONARY ====================

logger.info("ME: Saving dictionary") 

dictionary = corpora.Dictionary([word2id.keys()]) 
cPickle.dump(dictionary, open("dictionary/" + db_name + ".pkl", "w"))

#==================== TODO ==================
'''
1.	CC	Coordinating conjunction
2.	CD	Cardinal number
3.	DT	Determiner
4.	EX	Existential there
5.	FW	Foreign word
6.	IN	Preposition or subordinating conjunction
7.	JJ	Adjective
8.	JJR	Adjective, comparative
9.	JJS	Adjective, superlative
10.	LS	List item marker
11.	MD	Modal
12.	NN	Noun, singular or mass
13.	NNS	Noun, plural
14.	NNP	Proper noun, singular
15.	NNPS	Proper noun, plural
16.	PDT	Predeterminer
17.	POS	Possessive ending
18.	PRP	Personal pronoun
19.	PRP$	Possessive pronoun
20.	RB	Adverb
21.	RBR	Adverb, comparative
22.	RBS	Adverb, superlative
23.	RP	Particle
24.	SYM	Symbol
25.	TO	to
26.	UH	Interjection
27.	VB	Verb, base form
28.	VBD	Verb, past tense
29.	VBG	Verb, gerund or present participle
30.	VBN	Verb, past participle
31.	VBP	Verb, non-3rd person singular present
32.	VBZ	Verb, 3rd person singular present
33.	WDT	Wh-determiner
34.	WP	Wh-pronoun
35.	WP$	Possessive wh-pronoun
36.	WRB	Wh-adverb
'''