/**
 * Author: Peter Csiba
 * Email: peter.csiba@csip.sk 
 */
//!NOTE that some articles could be cited but they have no citations! 

#include "utils.hpp"

#define IS_SORT 0
#define IS_CITATIONS 1
#define MAX_BFS_OP 2000000
#define BFS_DEFAULT_DIST 100 

#define TO_COMPARE_COUNT 2408
#define DIST_FILE "data/to_compare.dat" 

map<int, int> pmid2index; 
map<int, int> index2pmid; 

vector<vector<double> > points; // N \times M array of points in LSI space (cosine similarity) 
vector<vector<int> > clusters = vector<vector<int> >(CLUSTER_COUNT, vector<int>());; 
vector<int> point2cluster; 

//graph of citations, both ways 
vector<vector<int> > citations; 

vector<vector<double> > dist_cos_clusters(CLUSTER_COUNT, vector<double>(CLUSTER_COUNT, 0.0)); 
//vector<vector<double> > dist_cit_clusters(CLUSTER_COUNT, vector<double>(CLUSTER_COUNT, 0.0)); 
//TODO citations 

void add_point_to_cluster(int point_index, int cluster_index){
  clusters[cluster_index].push_back(point_index); 
  point2cluster[point_index] = cluster_index; 
}

void make_max_value_cluster(){
  REP(i, points.size()) {
    add_point_to_cluster(i, max_index(points[i]));  
    if(i % 100000 == 0) cout << "   " << i << endl;
  }
}

void make_lsh_cluster(){
  string filename;
  cout << "make_lsh_cluster" << endl;
  
  vector<vector<int> > buckets(1 << LSH_BITS, vector<int>());
  vector<vector<double> > lsh_fns(LSH_BITS, vector<double>(TOPIC_COUNT, 0.0)); 
  
  REP(i, LSH_BITS) REP(j, TOPIC_COUNT) lsh_fns[i][j] = rand_double(-1.0, 1.0); 
  REP(i, points.size()) {
    int hash = 0; 
    REP(j, LSH_BITS) hash += (scalar(points[i], lsh_fns[j]) > 0) * (1 << j); 
    buckets[hash].push_back(i); 
    
    if(i % 10000 == 0) cout << "  " << i << endl;
  }
  
  vector<PD_II> to_sort; 
  
  cout << "printing out buckets" << endl;
  filename = "log/buckets"+ i2str(time(NULL)) +".dat";
  ofstream fbuckets(filename.c_str());
  REP(i, buckets.size()){
    double avg_sim = 0.0; 
    int n = buckets[i].size(); 
    
#if IS_SORT
    REP(a, n) REP(b, n) if(a < b){
	int p1 = buckets[i][a];
	int p2 = buckets[i][b];
	double cs = cos_sim(points[p1], points[p2]); 
	
	avg_sim += (cs) / ((double)(n*(n+1)/2)); 
	
	to_sort.push_back(PD_II(cs, PII(index2pmid[p1], index2pmid[p2]))); 
    }
#endif 
    
    fbuckets << buckets[i].size() << " " << avg_sim; 
    
    REP(j, buckets[i].size()) fbuckets << " " << buckets[i][j]; 
    fbuckets << endl;
    
    if(i % (1 << (LSH_BITS - 4)) == 0) cout << "  " << i << endl;
  }
  fbuckets.close(); 
  
#if IS_SORT
  //!NOTE that after the milion-th only the each hundred will be printed out 
  cout << "sorting" << endl;
  sort(to_sort.begin(), to_sort.end(), greater<PD_II>()); 
  cout << "  outputting closest" << endl; 
  filename = "log/dist_" + i2str(time(NULL)) + ".dat"; 
  ofstream fdist(filename.c_str());  
  REP(i, to_sort.size()){
    if(i > 1000000 && i % 100 == 0) continue; 
    
    fdist << to_sort[i].first << "\t" << to_sort[i].second.first << "\t" << to_sort[i].second.second << endl;
    
    if(i % (to_sort.size() / 100) == 0) cout << "    " << i << endl;
  } 
#endif
}

void compute_avg_cos_dist_between_clusters(){
  cout << "compute_avg_cos_dist_between_clusters" << endl;
  
  int counts[CLUSTER_COUNT][CLUSTER_COUNT];
  REP(i, CLUSTER_COUNT) REP(j, CLUSTER_COUNT) dist_cos_clusters[i][j] = 0.0; 
  
  REP(i, COUNT_CLUSTER_DISTANCE_SAMPLES){
    int p1 = rand() % points.size(); 
    int p2 = rand() % points.size(); 
    double d = cos_sim(points[p1], points[p2]); 
    
    int c1 = point2cluster[p1];
    int c2 = point2cluster[p2];
    dist_cos_clusters[c1][c2] += d;
    counts[c1][c2] += 1; 
    
    if(i % 100000 == 0) cout << "  " << i << endl;
  }
  
  REP(i, CLUSTER_COUNT) REP(j, CLUSTER_COUNT) dist_cos_clusters[i][j] /= (double) counts[i][j]; 
}

void compute_avg_cit_dist_between_clusters(){
 
}

void evaluate(){
  cout << "Evaluating clusters" << endl;
  
  cout << "  Distances" << endl;
  REP(i, 10) {
    cout << "    "; 
    REP(j, 10) cout << dist_cos_clusters[i][j] << " "; 
    cout << endl; 
  }
  
  cout << "  Samples" << endl; 
  REP(i, 10) {
    cout << "    Cluster " << i << " size=" << clusters[i].size(); 
    REP(j, 8) cout << "  " << index2pmid[clusters[i][j]]; 
    cout << endl;
  }
}

void print_cos_sim(int index1, int index2){
  cout << index1 << "," << index2 << cos_sim(points[pmid2index[index1]], points[pmid2index[index2]]) << endl;
}

int con_bfs(int from, int to, int max_op = MAX_BFS_OP){
  cout << "con_bfs(" << from << "," << to << ")" << endl;
  
  set<int> V; 
  queue<PII> N;
  N.push(PII(0, from));
  
  int op = 0; 
  
  while(!N.empty()){
    int d = N.front().first;
    int n = N.front().second;
    if(n == to) return d; 
    N.pop(); 
    
    if(V.find(n) != V.end()) continue; 
    V.insert(n);  
    
    REP(i, citations[n].size()) {
      if(op >= max_op){
	continue; 
      }
      N.push(PII(d+1, citations[n][i])); 
      op++;
      
      if(op % (max_op / 10) == 0) cout << "  op=" << op << endl;
    }
  }
 
  return BFS_DEFAULT_DIST; 
}

void compare_dist_cos_cit(){
  cout << "compare_dist_cos_cit" << endl;
  cout << "  loading to compare file=" << DIST_FILE << endl;
  
  ifstream fdist(DIST_FILE); 
  vector<PD_II> to_compare; 
  REP(i, TO_COMPARE_COUNT) {
    PD_II p; 
    fdist >> p.first >> p.second.first >> p.second.second; 
    p.second.first = pmid2index[p.second.first];
    p.second.second = pmid2index[p.second.second];
    
    to_compare.push_back(p); 
  }

  cout << "  calculating citation wal distance" << endl;
  
  string filename = "log/compare"+ i2str(time(NULL)) +".dat";
  ofstream fcompare(filename.c_str());
  for(int i=1000; i<to_compare.size() ; i++){
    PII p = to_compare[i].second; 
    int d = con_bfs(p.first, p.second); 
    fcompare << to_compare[i].first << "\t" << d << "\t" << p.first << "\t" << p.second << endl;
    cout << "  i=" << i << ":" << to_compare[i].first << "\t" << d << "\t" << p.first << "\t" << p.second << endl;
  }
  fcompare.close(); 
}

int main(int argc, char* argv[]){
  srandom(time(NULL)); 
  
  REP(i, 10) cout << rand_double(-1.0, 1.0) << endl; 
  
  cout << "Loading vectors file" << endl;
  
  int N; 
  cin >> N >> TOPIC_COUNT; 
  point2cluster.resize(N, -1);
  
  cout << "  Size:" << N << "x" << TOPIC_COUNT << endl;
  
  REP(i, N){
    int pmid;
    cin >> pmid;
    pmid2index[pmid] = i;
    index2pmid[i] = pmid; 
    citations.push_back(vector<int>()); 
    
    points.push_back(vector<double>(TOPIC_COUNT, 0.0)); 
    REP(j, TOPIC_COUNT) cin >> points[i][j];
    
    if(i % 40000 == 0) cout << "  " << i << endl;
  }
  
#if IS_CITATIONS
  cout << "Loading citations" << endl; 
  ifstream fcitations("citations.csv"); 
  string line;
  int cc = 0; 
  int skipped = 0; 
  unsigned int max_deg = 0; 
  while(getline(fcitations, line)){
    vector<int> vint = read_csv_int_line(line); 
    line.clear(); 
    
    for(int i=1; i<vint.size() ; i++){
      if(pmid2index.count(vint[0]) == 0 || pmid2index.count(vint[i]) == 0){
	skipped++; 
	continue; 
      }
      
      int index1 = pmid2index[vint[0]];
      int index2 = pmid2index[vint[i]]; 
      citations[index1].push_back(index2); 
      citations[index2].push_back(index1); 
      
      if(cc++ % 100000 == 0 && i <= 2) cout << "  " << cc << "(" << index1 << "," << index2 << ") orig(" << vint[0] << "," << vint[i] << ")" << endl;
    }
    
    max_deg = max(max_deg, (unsigned int) vint.size()); 
  }
  cout << "  max_deg=" << max_deg << endl;
  cout << "  skipped=" << skipped << endl;
  
  compare_dist_cos_cit(); 
  return 0; 
#endif 
  
  cout << "Makeing clusters" << endl;
  //ake_max_value_cluster(); 
  //compute_avg_cos_dist_between_clusters();
  //evaluate(); 
  
  make_lsh_cluster(); 
  //compute_avg_cos_dist_between_clusters();
  //evaluate(); 
  
}

void cos_sim_test(){
  // looks good 
  print_cos_sim(18648627,18648562);
  print_cos_sim(18647861,18647860);
  print_cos_sim(18682795,18682764);
  print_cos_sim(18668757,18668742);
  print_cos_sim(18663132,18663131);
  print_cos_sim(18658223,18658222);
  print_cos_sim(18629219,18629157);
  print_cos_sim(18613949,18613946);
  print_cos_sim(18668204,18626576);
  print_cos_sim(18629158,18628849);
  print_cos_sim(18629123,18629059);
  print_cos_sim(18648582,18648573);
  print_cos_sim(18648603,18648571);
  print_cos_sim(18668273,18611266);
  print_cos_sim(18625747,18606673);
  print_cos_sim(18617608,18617607);
  print_cos_sim(18606854,18606850);
  print_cos_sim(18641471,18641467);
}