#Author: Peter Csiba
#Email: petherz@gmail.com 
#Company: Black Swan Rational 

#TODO move to group_similar.py (via cPickle save/load(lsi) 
#TODO query_radius()

#TODO consider dimension reduction 
#TODO just output cosine distance and do the KNN with C++ or something so much faster 

#TODO Efficient K-Nearest Neighbor Graph Construction for Generic Similarity Measures
#TODO Similarity Learning for Nearest Neighbor Classification
#TODO FLANN

import logging
import cPickle
import csv
import sys 
import math 
import time

from scipy import spatial

EACH_PAIR = 0
ONLY_CITED = 1 # calculate distance only between citation links 
MEGA_MOTAC_KNN = 2 # TODO 
TO_CONSIDER = ONLY_CITED

# cosine distance is something between 0.02 and 1.22, the less the closer 
CITATION_DISTANCE_DEFAULT = 1.0 
CITATION_DISTANCE_WEIGHT = 0.3
CITATION_MAX_LOG = math.log(10887.0 + 1.0, 2.0) 

#========== LOGGING ============

logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.INFO)
logger = logging.getLogger('nearest') 

corpus_name = sys.argv[1] 
topic_model_name = sys.argv[2] 
NUM_TOPICS = int(sys.argv[3])  #TODO there must be a way to calculate it from LSI model 

run_id = corpus_name + "_" + str(int(time.time()))

#=================== DICTIONARY ==============
logger.info("ME: Loading dictionary") 
dictionary = cPickle.load(open("dictionary/" + topic_model_name + ".pkl", "r")) 
logger.info("ME: Dictionary("+str(len(dictionary))+")") 

#=========== LOAD LSI ===========
logger.info("ME: Loading LSI directly") 
lsi = cPickle.load(open("model/lsi_" + topic_model_name + ".pkl", "r"))
logger.info("ME:   Done") 

#==================== LOAD CORPUS DATA -> LSI VECTOR======================
logger.info("ME: Loading doc to RAM (via doc2bow and lsi transformers)")
corpus_csv = csv.reader(open("model/docs_used_" + corpus_name + ".csv", "r")) 
corpus_data = {} #map id to abstract tokens 
cc = 0 
for line in corpus_csv: 
  #convert sparse representation to normal vector 
  row = [0.0] * NUM_TOPICS 
  for sparse_el in lsi[dictionary.doc2bow(line[1:])]: 
    row[sparse_el[0]] = sparse_el[1]
  
  corpus_data[line[0]] = row 
  
  cc += 1
  if cc % 40000 == 0:
    print logger.info("ME:   " + str(cc))
  
  #print "line=", line[1:]
  #print "bow", dictionary.doc2bow(line[1:])
  #print "row", row

#=========== LOAD & COUNT CITATIONS =============
logger.info("ME: Loading citations") 
cit_csv = csv.reader(open("citations.csv", "r"))

in_cit = {} # int(key) -> float( count of incomming citations ) 
out_cit = {} # int(key) -> set( outgoing citations ) 

cc = 0 
for cit in cit_csv: 
  for ic in cit[1:]:
    #consider only articles in corpus 
    if ic in corpus_data:
      if not ic in in_cit: 
	in_cit[ic] = 0.0
      in_cit[ic] += 1.0 
  
  if cit[0] in corpus_data:
    out_cit[cit[0]] = set(cit[1:]) 
  
  cc += 1
  if cc % 100000 == 0:
    logger.info("ME:   " + str(cc))
    
#=============== CITATION DISTANCE ===================
# @precondition id1 in in_cit[id2] 
def unsafe_get_citation_distance(id1, id2): 
  return math.log(1.0 + in_cit[id2], 2.0) / CITATION_MAX_LOG

#returns distance (0.0, 1.0] 
def get_citation_distance(id1, id2):
  id1_cites_id2 = id2 in out_cit[id1]
  id2_cites_id1 = id1 in out_cit[id2]
  
  if id1_cites_id2: 
    return unsafe_get_citation_distance(id1, id2)
  if id2_cites_id1: 
    return unsafe_get_citation_distance(id2, id1)
  
  return CITATION_DISTANCE_DEFAULT 
    
#=================== CALCULATE PAIRWASE DISTANCES =================
logger.info("ME: Finding closest documents")
dist_doc = []
cc = 0 
#================ DISTANCE BETWEEN EACH PAIR ====================

if TO_CONSIDER == EACH_PAIR: 
  logger.info("ME:   Considering each pair ")
  for k1,p1 in corpus_data.iteritems(): 
    for k2,p2 in corpus_data.iteritems(): 
      if k2 >= k1:
	continue 
      
      #The resulting similarity ranges from -1 meaning exactly opposite, to 1 meaning exactly the same, with 0 usually indicating independence, and in-between values indicating intermediate similarity or dissimilarity.
      dcos = spatial.distance.cosine(p1, p2)
      dcit = CITATION_DISTANCE_WEIGHT * get_citation_distance(k1, k2) 
      dist_doc.append([dcos + dcit, k1, k2, dcos, dcit]) 
    
    cc += 1
    if cc % 100 == 0: 
      logger.info("ME:     " + str(cc)) 
  
#================ DISTANCE BETWEEN CITATION LINKS ====================

if TO_CONSIDER == ONLY_CITED: 
  logger.info("ME:   Considering only cited pairs")
  cc = 0 
  for kout, outgoing in out_cit.iteritems(): 
    for kin in outgoing: 
      #!DEVELOPER # when using smaller corpus for tests 
      if not kout in corpus_data or not kin in corpus_data:
	continue
	  
      if kout >= kin: #TODO inconsistent data (especially the case ==) 
	continue
	  
      p1 = corpus_data[kout] 
      p2 = corpus_data[kin] 
      dcos = spatial.distance.cosine(p1, p2) 
      dcit = CITATION_DISTANCE_WEIGHT * unsafe_get_citation_distance(kout, kin)
      dist_doc.append([dcos + dcit, kout, kin, dcos, dcit]) 
      
      cc += 1
      if cc % 100000 == 0: 
	logger.info("ME:   " + str(cc)) 

#================== SORT & PRINT DISTANCES ===================
if TO_CONSIDER == EACH_PAIR or TO_CONSIDER == ONLY_CITED:
  logger.info("ME: Sorting")
  dist_doc.sort()

  print "Closest articles (begin & end 10000): "
  dist_csv = csv.writer(open("log/dist_" + run_id + ".csv", "w"))
  cd = 0 
  for line in dist_doc: 
    cd += 1
    
    if cd <= 10000 or (len(dist_doc) - 10000) <= cd: 
      dist_csv.writerow(line) 
    
    if cd <= 30 or (len(dist_doc) - 30) <= cd: 
      print "\t".join([str(x) for x in line]) 

#====================== FANCY KNN ==================

if TO_CONSIDER == MEGA_MOTAC_KNN:
  logger.info("ME: Fancy KNN / LSH")
  logger.info("ME: TODO")
  

'''
logger.info("ME: Finding similar articles") 
# !WARNING - could be very big 
lsi_docs = []
for doc_bow in corpus_memory_friendly:
  #convert sparse representation to normal vector 
  row = [0.0] * NUM_TOPICS 
  for sparse_el in lsi[doc_bow]: 
    row[sparse_el[0]] = sparse_el[1]
  lsi_docs.append(row) 
  
logger.info("ME: Constructing BallTree")
ball_tree = BallTree(lsi_docs, leaf_size=10)         

logger.info("ME: Finding closest documents")
dist_doc = [] 
for i in xrange(0,len(lsi_docs)): 
  dist, ind = ball_tree.query(lsi_docs[i], k=10)  
  #print "dist='",dist[0],"'" 
  #print "ind='",ind[0],"'" 
  
  for b in xrange(0, len(dist[0])): 
    #add it only from one side (and not dist(i,i)) 
    if i < ind[0][b]: 
      dist_doc.append([dist[0][b], i, ind[0][b]]) 
      
  if (i+1) % 1000 == 0:
    logger.info("ME:   " + str(i+1)) 

logger.info("ME: Sorting")
dist_doc.sort()

print "Closest articles (some may missing): "
dist_csv = csv.writer(open("log/dist_" + run_id + ".csv", "w"))
cd = 0 
for line in dist_doc: 
  dist_csv.writerow(line) 
  
  cd += 1
  if cd <= 50: 
    print line[0], "\t", line[1], "\t", line[2]
'''


#TODO clustering / categorizing articles 