import xml.etree.cElementTree as ET
import sys
import time
import datetime
import re

def get_indexdate(xml):
  aid = xml.iter("PubMedPubDate")
  for x in aid:
    if "PubStatus" in x.attrib and x.attrib["PubStatus"] == "entrez":
      year = x.find("Year").text
      monthe = x.find("Month")
      if monthe is not None:
        month = monthe.text
      else:
        month = "12"
        fail_month += 1
      daye = x.find("Day")
      if daye is not None:
        day = daye.text
      else:
        day = "28"
        fail_day += 1
      date = "%s-%s-%s" % (year, month, day)
      return date
        
  return None

def get_doi(xml):
  aid = xml.iter("ArticleId")
  for x in aid:
    if "IdType" in x.attrib and x.attrib["IdType"] == "doi":
      return x.text
  return None

def get_volume(xml):
  volume_list = list(xml.iter("Volume"))
  if len(volume_list) > 0:
    return volume_list[0].text
  return None

def get_issue(xml):
  issue_list = list(xml.iter("Issue"))
  if len(issue_list) > 0:
    return issue_list[0].text
  return None

def get_pgn(xml):
  pgn_list = list(xml.iter("MedlinePgn"))
  if len(pgn_list) > 0:
    return pgn_list[0].text
  return None

def parse_article(article):
  xml = ET.fromstring(article)
  abstract_list = list(xml.iter("AbstractText"))
  title_list = list(xml.iter("ArticleTitle"))
  if len(title_list) < 1:
    return "", "", "", "", "", "", "", "" 

  abstract = ""
  for x in abstract_list:
    abstract += " "
    if "Label" in x.attrib and x.attrib["Label"] != "UNLABELLED":
      abstract += x.attrib["Label"] + ": "
    if x.text:
      abstract += x.text
    else:
      pass
#      print ET.tostring(x)

  pubdate_list = list(xml.iter("PubDate"))
  if pubdate_list[0].find("Year") != None:
    year = pubdate_list[0].find("Year").text
    if pubdate_list[0].find("Month") != None:
      month = pubdate_list[0].find("Month").text  
    else:
      month = "Jan"
    if pubdate_list[0].find("Day") != None:
      day = pubdate_list[0].find("Day").text  
    else:
      day = "1"
    try:
      date = time.strptime("%s %s %s" % (day, month, year), "%d %b %Y") 
    except:
      try:
        date = time.strptime("%s %s" % (month, year), "%b %Y") 
      except:  
        date = time.strptime("%s %s %s" % (day, month, year), "%d %m %Y")  

  else:
    res = re.findall(r"\d\d\d\d", ET.tostring(pubdate_list[0])) 
    if len(res) > 0:
      date = time.strptime(res[0], "%Y")
    else:
      date = time.time()

  authors = []
  author_list = list(xml.iter("AuthorList"))
  if len(author_list) > 0:
    for author in author_list[0]:
      name = ""
      for x in author:
        if x.tag == "Initials":
          continue
        if not x.text:
          continue
        if len(name) > 0:
          name += " "
        name += x.text
      if len(name) > 1:
        authors.append(name)

  issn_list = list(xml.iter("ISSN"))
  issn = None
  if len(issn_list) > 0:
    issn = issn_list[0].text

  journal_list = list(xml.iter("Title"))
  journal = None
  if len(journal_list) > 0:
    journal = journal_list[0].text

  mesh_list = [x.text for x in xml.iter("DescriptorName")]

  other_data = {}
  indexdate = get_indexdate(xml)
  if indexdate:
    other_data["indexdate"] = indexdate
  doi = get_doi(xml)
  if doi:
    other_data["doi"] = doi
  volume = get_volume(xml)
  if volume:
    other_data["volume"] = volume
  issue = get_issue(xml)
  if issue:
    other_data["issue"] = issue
  pgn = get_pgn(xml)
  if pgn:
    other_data["pgn"] = pgn

  return title_list[0].text, abstract, date, authors, issn, journal, mesh_list, other_data

